import datetime
from typing import Literal

from pydantic import BaseModel, Field


class ResponseModel(BaseModel):
    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema, handler):
        # Don't clutter response model docs
        schema = handler(core_schema)
        schema = handler.resolve_ref_schema(schema)
        schema.pop('required', None)
        return schema


class NotFoundResponse(ResponseModel):
    detail: str = Field("Error details")


class Source(ResponseModel):
    id: int = Field(
        description="Bright Sky source ID",
        examples=[
            6007,
        ],
    )
    dwd_station_id: str = Field(
        description="DWD weather station ID",
        examples=[
            "01766",
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wmo_station_id: str = Field(
        description="WMO weather station ID",
        examples=[
            "10315",
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    station_name: str = Field(
        description="DWD weather station name",
        examples=[
            "Münster/Osnabrück",
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    observation_type: Literal[
        'historical',
        'current',
        'synop',
        'forecast',
    ] = Field(
        description="Source type",
    )
    first_record: datetime.datetime = Field(
        description="Timestamp of first available record for this source",
        examples=[
            "2010-01-01T00:00+02:00",
        ],
    )
    last_record: datetime.datetime = Field(
        description="Timestamp of latest available record for this source",
        examples=[
            "2023-08-07T12:40+02:00",
        ],
    )
    lat: float = Field(
        description="Station latitude, in decimal degrees",
        examples=[
            52.1344,
        ],
    )
    lon: float = Field(
        description="Station longitude, in decimal degrees",
        examples=[
            7.6969,
        ],
    )
    height: float = Field(
        description="Station height, in meters",
        examples=[
            47.8,
        ],
    )
    distance: int = Field(
        default=None,
        description="Distance of weather station to the requested `lat` and `lon` (if given), in meters",  # noqa
        examples=[
            16365,
        ],
    )


class BaseWeatherRecord(ResponseModel):
    timestamp: datetime.datetime = Field(
        description="ISO 8601-formatted timestamp of this weather record",
        examples=[
            "2023-08-07T12:30:00+00:00",
        ],
    )
    source_id: int = Field(
        description="Bright Sky source ID for this record",
        examples=[
            238685,
        ],
    )
    cloud_cover: float = Field(
        description="Total cloud cover at timestamp",
        examples=[
            12.1,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    condition: Literal[
        "dry",
        "fog",
        "rain",
        "sleet",
        "snow",
        "hail",
        "thunderstorm",
    ] = Field(
        description="Current weather conditions. Unlike the numerical parameters, this field is not taken as-is from the raw data (because it does not exist), but is calculated from different fields in the raw data as a best effort. Not all values are available for all source types.",  # noqa
        json_schema_extra={
            'nullable': True,
        },
    )
    dew_point: float = Field(
        description="Dew point at timestamp, 2 m above ground",
        examples=[
            -2.5,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    icon: Literal[
        "clear-day",
        "clear-night",
        "partly-cloudy-day",
        "partly-cloudy-night",
        "cloudy",
        "fog",
        "wind",
        "rain",
        "sleet",
        "snow",
        "hail",
        "thunderstorm",
    ] = Field(
        description="Icon alias suitable for the current weather conditions. Unlike the numerical parameters, this field is not taken as-is from the raw data (because it does not exist), but is calculated from different fields in the raw data as a best effort. Not all values are available for all source types.",  # noqa
        json_schema_extra={
            'nullable': True,
        },
    )
    pressure_msl: float = Field(
        description="Atmospheric pressure at timestamp, reduced to mean sea level",  # noqa
        examples=[
            1015.1,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    relative_humidity: int = Field(
        description="Relative humidity at timestamp",
        examples=[
            40,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    temperature: float = Field(
        description="Air temperature at timestamp, 2 m above the ground",
        examples=[
            10.6,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    visibility: int = Field(
        description="Visibility at timestamp",
        examples=[
            50000,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    fallback_source_ids: dict = Field(
        description="Object mapping meteorological parameters to the source IDs of alternative sources that were used to fill up missing values in the main source",  # noqa
        examples=[
            {"pressure_msl": 11831, "wind_speed_30": 11831},
        ],
    )


class WeatherRecord(BaseWeatherRecord):
    precipitation: float = Field(
        description="Total precipitation during previous 60 minutes",
        examples=[
            1.8,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    solar: float = Field(
        description="Solar irradiation during previous 60 minutes",
        examples=[
            0.563,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    sunshine: int = Field(
        description="Sunshine duration during previous 60 minutes",
        examples=[
            58,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_direction: int = Field(
        description="Mean wind direction during previous hour, 10 m above the ground",  # noqa
        examples=[
            70,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_speed: float = Field(
        description="Mean wind speed during previous hour, 10 m above the ground",  # noqa
        examples=[
            12.6,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_direction: int = Field(
        description="Direction of maximum wind gust during previous hour, 10 m above the ground",  # noqa
        examples=[
            50,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_speed: float = Field(
        description="Speed of maximum wind gust during previous hour, 10 m above the ground",  # noqa
        examples=[
            33.5,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    precipitation_probability: int = Field(
        description="Probability of more than 0.1 mm of precipitation in the previous hour (only available in forecasts)",  # noqa
        examples=[
            46,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    precipitation_probability_6h: int = Field(
        description="Probability of more than 0.2 mm of precipitation in the previous 6 hours (only available in forecasts at 0:00, 6:00, 12:00, and 18:00 UTC)",  # noqa
        examples=[
            75,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )


class CurrentWeatherRecord(BaseWeatherRecord):
    precipitation_10: float = Field(
        description="Total precipitation during previous 10 minutes",
        examples=[
            0.8,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    precipitation_30: float = Field(
        description="Total precipitation during previous 30 minutes",
        examples=[
            1.2,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    precipitation_60: float = Field(
        description="Total precipitation during previous 60 minutes",
        examples=[
            1.8,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    solar_10: float = Field(
        description="Solar irradiation during previous 10 minutes",
        examples=[
            0.081,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    solar_30: float = Field(
        description="Solar irradiation during previous 30 minutes",
        examples=[
            0.207,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    solar_60: float = Field(
        description="Solar irradiation during previous 60 minutes",
        examples=[
            0.48,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    sunshine_30: int = Field(
        description="Sunshine duration during previous 30 minutes",
        examples=[
            28,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    sunshine_60: int = Field(
        description="Sunshine duration during previous 60 minutes",
        examples=[
            52,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_direction_10: int = Field(
        description="Mean wind direction during previous 10 minutes, 10 m above the ground",  # noqa
        examples=[
            70,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_direction_30: int = Field(
        description="Mean wind direction during previous 30 minutes, 10 m above the ground",  # noqa
        examples=[
            70,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_direction_60: int = Field(
        description="Mean wind direction during previous 60 minutes, 10 m above the ground",  # noqa
        examples=[
            70,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_speed_10: float = Field(
        description="Mean wind speed during previous 10 minutes, 10 m above the ground",  # noqa
        examples=[
            12.6,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_speed_30: float = Field(
        description="Mean wind speed during previous 30 minutes, 10 m above the ground",  # noqa
        examples=[
            12.6,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_speed_60: float = Field(
        description="Mean wind speed during previous 60 minutes, 10 m above the ground",  # noqa
        examples=[
            12.6,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_direction_10: int = Field(
        description="Direction of maximum wind gust during previous 10 minutes, 10 m above the ground",  # noqa
        examples=[
            50,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_direction_30: int = Field(
        description="Direction of maximum wind gust during previous 30 minutes, 10 m above the ground",  # noqa
        examples=[
            50,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_direction_60: int = Field(
        description="Direction of maximum wind gust during previous 60 minutes, 10 m above the ground",  # noqa
        examples=[
            50,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_speed_10: float = Field(
        description="Speed of maximum wind gust during previous 10 minutes, 10 m above the ground",  # noqa
        examples=[
            33.5,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_speed_30: float = Field(
        description="Speed of maximum wind gust during previous 30 minutes, 10 m above the ground",  # noqa
        examples=[
            33.5,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    wind_gust_speed_60: float = Field(
        description="Speed of maximum wind gust during previous 60 minutes, 10 m above the ground",  # noqa
        examples=[
            33.5,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )


class SynopRecord(CurrentWeatherRecord):
    sunshine_10: int = Field(
        description="Sunshine duration during previous 10 minutes",
        examples=[
            8,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )


class SourcesResponse(ResponseModel):
    sources: list[Source]


class CurrentWeatherResponse(ResponseModel):
    weather: CurrentWeatherRecord
    sources: list[Source]


class WeatherResponse(ResponseModel):
    weather: list[WeatherRecord]
    sources: list[Source]


class SynopResponse(ResponseModel):
    weather: list[SynopRecord]
    sources: list[Source]


class RadarRecord(ResponseModel):
    timestamp: datetime.datetime = Field(
        description="ISO 8601-formatted timestamp of this radar record",
        examples=[
            "2023-08-07T08:00:00+00:00",
        ],
    )
    source: str = Field(
        description="Unique identifier for DWD radar product source of this radar record",  # noqa
        examples=[
            "RADOLAN::RV::2023-08-08T11:45:00+00:00",
        ],
    )
    precipitation_5: str = Field(
        description="Pixelwise 5-minute precipitation data, in units of 0.01 mm / 5 min. Depending on the `format` parameter, this field contains either a two-dimensional array of integers (`plain`), or a base64 string (`bytes` or `compressed`).",  # noqa
        examples=[
            "eF5jGAWjYBTQEQAAA3IAAQ==",
        ],
    )


class RadarResponse(ResponseModel):
    radar: list[RadarRecord]
    geometry: dict = Field(
        description="GeoJSON-formatted bounding box of returned radar data, i.e. lat/lon coordinates of the four corners.",  # noqa
        examples=[
            {
                "type": "Polygon",
                "coordinates": [
                    [7.44365, 52.08712],
                    [7.45668, 51.90644],
                    [7.7487, 51.914],
                    [7.73716, 52.09473],
                ],
            },
        ],
    )
    bbox: list[int] = Field(
        description="Bounding box (top, left, bottom, right) calculated from the supplied position and distance. Only returned if you supplied `lat` and `lon`.",  # noqa
        examples=[
            [100, 100, 300, 300],
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    latlon_position: dict = Field(
        description="Exact x-y-position of the supplied position. Only returned if you supplied `lat` and `lon`.",  # noqa
        examples=[
            {"x": 10.244, "y": 10.088},
        ],
        json_schema_extra={
            'nullable': True,
        },
    )


class Alert(ResponseModel):
    id: int = Field(
        description="Bright Sky-internal ID for this alert",
        examples=[
            279977,
        ],
    )
    alert_id: str = Field(
        description="Unique CAP message identifier",
        examples=[
            "2.49.0.0.276.0.DWD.PVW.1691344680000.2cf9fad6-dc83-44ba-9e88-f2827439da59",  # noqa
        ],
    )
    status: Literal["actual", "test"] = Field(
        description="Alert status",
    )
    effective: datetime.datetime = Field(
        description="Alert issue time",
        examples=[
            "2023-08-06T17:58:00+00:00",
        ],
    )
    onset: datetime.datetime = Field(
        description="Expected event begin time",
        examples=[
            "2023-08-07T08:00:00+00:00",
        ],
    )
    expires: datetime.datetime = Field(
        description="Expected event end time",
        examples=[
            "2023-08-07T19:00:00+00:00",
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    category: Literal["met", "health"] = Field(
        description="Alert category, meteorological message (`met`) or public health related message (`health`)",  # noqa
        json_schema_extra={
            'nullable': True,
        },
    )
    response_type: Literal["prepare", "allclear", "none", "monitor"] = Field(
        description="Code denoting type of action recommended for target audience",  # noqa
        json_schema_extra={
            'nullable': True,
        },
    )
    urgency: Literal["immediate", "future"] = Field(
        description="Alert time frame",
        json_schema_extra={
            'nullable': True,
        },
    )
    severity: Literal["minor", "moderate", "severe", "extreme"] = Field(
        description="Alert severity",
        json_schema_extra={
            'nullable': True,
        },
    )
    certainty: Literal["observed", "likely"] = Field(
        description="Alert certainty",
        json_schema_extra={
            'nullable': True,
        },
    )
    event_code: int = Field(
        description="DWD event code",
        examples=[
            51,
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    event_en: str = Field(
        description="Label for DWD event code (English)",
        examples=[
            "wind gusts",
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    event_de: str = Field(
        description="Label for DWD event code (German)",
        examples=[
            "WINDBÖEN",
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    headline_en: str = Field(
        description="Alert headline (English)",
        examples=[
            "Official WARNING of WIND GUSTS",
        ],
    )
    headline_de: str = Field(
        description="Alert headline (German)",
        examples=[
            "Amtliche WARNUNG vor WINDBÖEN",
        ],
    )
    description_en: str = Field(
        description="Alert description (English)",
        examples=[
            "There is a risk of wind gusts (level 1 of 4).\nMax. gusts: 50-60 km/h; Wind direction: west; Increased gusts: near showers and in exposed locations < 70 km/h",  # noqa
        ],
    )
    description_de: str = Field(
        description="Alert description (German)",
        examples=[
            "Es treten Windböen mit Geschwindigkeiten zwischen 50 km/h (14 m/s, 28 kn, Bft 7) und 60 km/h (17 m/s, 33 kn, Bft 7) aus westlicher Richtung auf. In Schauernähe sowie in exponierten Lagen muss mit Sturmböen bis 70 km/h (20 m/s, 38 kn, Bft 8) gerechnet werden.",  # noqa
        ],
    )
    instruction_en: str = Field(
        description="Additional instructions and safety advice (English)",
        examples=[
            "NOTE: Be aware of the following possible dangers: The downpours can cause temporary traffic disruption.",  # noqa
        ],
        json_schema_extra={
            'nullable': True,
        },
    )
    instruction_de: str = Field(
        description="Additional instructions and safety advice (German)",
        examples=[
            "ACHTUNG! Hinweis auf mögliche Gefahren: Während des Platzregens sind kurzzeitig Verkehrsbehinderungen möglich.",  # noqa
        ],
        json_schema_extra={
            'nullable': True,
        },
    )


class Location(ResponseModel):
    warn_cell_id: int = Field(
        description="Municipality warn cell ID of given location",
        examples=[
            803159016,
        ],
    )
    name: str = Field(
        description="Municipality name",
        examples=[
            "Stadt Göttingen",
        ],
    )
    name_short: str = Field(
        description="Shortened municipality name",
        examples=[
            "Göttingen",
        ],
    )
    district: str = Field(
        description="District name",
        examples=[
            "Göttingen",
        ],
    )
    state: str = Field(
        description="Federal state name",
        examples=[
            "Niedersachsen",
        ],
    )
    state_short: str = Field(
        description="Shortened federal state name",
        examples=[
            "NI",
        ],
    )


class AlertsResponse(ResponseModel):
    alerts: list[Alert]
    location: Location = Field(
        json_schema_extra={
            'nullable': True,
        },
    )



import base64
import contextlib
from pathlib import Path
from typing import Any, Annotated

import asyncpg
import orjson
from fastapi import FastAPI, HTTPException, Query, Request, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import ORJSONResponse

import brightsky
from brightsky import query
from brightsky.enhancements import enhance
from brightsky.settings import settings

from .models import (
    AlertsResponse,
    CurrentWeatherResponse,
    NotFoundResponse,
    RadarResponse,
    SourcesResponse,
    SynopResponse,
    WeatherResponse,
)
from .params import (
    AlertsParams,
    CurrentWeatherParams,
    RadarParams,
    SourcesParams,
    SynopParams,
    WeatherParams,
)


with open(Path(__file__).parent / 'intro.md') as f:
    DESCRIPTION = f.read()


OPENAPI_METADATA = {
    'title': 'Bright Sky',
    'description': DESCRIPTION,
    'version': brightsky.__version__,
    'contact': {
        'name': 'Jakob de Maeyer',
        'email': 'jakob@brightsky.dev',
        'url': 'https://brightsky.dev/',
    },
    'license_info': {
        'name': 'MIT License',
        'identifier': 'MIT',
    },
    'servers': [
        {'url': settings.SERVER_URL},
    ],
}


ctx = {}


async def init_connection(conn):
    await conn.set_type_codec(
        'real',
        encoder=str,
        decoder=float,
        schema='pg_catalog',
    )


@contextlib.asynccontextmanager
async def lifespan(app: FastAPI):
    async with asyncpg.create_pool(
        dsn=settings.DATABASE_URL,
        min_size=1,
        max_size=settings.DATABASE_CONNECTION_POOL_SIZE,
        init=init_connection,
    ) as pool:
        ctx['pool'] = pool
        yield
        del ctx['pool']


def make_app():
    app = FastAPI(
        lifespan=lifespan,
        **OPENAPI_METADATA,
    )
    app.add_middleware(
        CORSMiddleware,
        allow_origins=(
            ["*"]
            if settings.CORS_ALLOW_ALL_ORIGINS
            else settings.CORS_ALLOWED_ORIGINS
        ),
        allow_headers=(
            ["*"]
            if settings.CORS_ALLOW_ALL_HEADERS
            else settings.CORS_ALLOWED_HEADERS
        ),
    )
    return app


app = make_app()


@app.exception_handler(query.NoData)
async def not_found(request, exc) -> NotFoundResponse:
    return ORJSONResponse(
        status_code=404,
        content={'detail': str(exc)},
    )


common_responses = {
    404: {'model': NotFoundResponse},
}


@app.get('/', include_in_schema=False)
async def status():
    return ORJSONResponse({
        'name': 'brightsky',
        'version': brightsky.__version__,
        'status': 'ok',
    })


@app.get(
    '/sources',
    operation_id='getSources',
    summary='Weather sources (stations)',
    tags=['Internals'],
    responses=common_responses,
)
async def sources(
    q: Annotated[SourcesParams, Query()],
) -> SourcesResponse:
    """
    Return a list of all Bright Sky sources matching the given location
    criteria, ordered by distance.

    You must supply both `lat` and `lon` _or_ one of `dwd_station_id`,
    `wmo_station_id`, or `source_id`.
    """
    result = await query.sources(
        ctx['pool'],
        lat=q.lat,
        lon=q.lon,
        max_dist=q.max_dist,
        dwd_station_ids=q.dwd_station_ids,
        wmo_station_ids=q.wmo_station_ids,
        source_ids=q.source_ids,
        ignore_type=True,
    )
    enhance(result, timezone=q.timezone)
    return ORJSONResponse(result)


@app.get(
    '/current_weather',
    operation_id='getCurrentWeather',
    summary='Current weather',
    responses=common_responses,
)
async def current_weather(
    q: Annotated[CurrentWeatherParams, Query()],
) -> CurrentWeatherResponse:
    """
    Returns current weather for a given location.

    To set the location for which to retrieve weather, you must supply both
    `lat` and `lon` _or_ one of `dwd_station_id`, `wmo_station_id`, or
    `source_id`.

    This endpoint is different from the other weather endpoints in that it does
    not directly correspond to any of the data available from the DWD Open Data
    server. Instead, it is a best-effort solution to reflect current weather
    conditions by compiling [SYNOP observations](/operations/getSynop) from the
    past one and a half hours.
    """
    result = await query.current_weather(
        ctx['pool'],
        lat=q.lat,
        lon=q.lon,
        max_dist=q.max_dist,
        dwd_station_ids=q.dwd_station_ids,
        wmo_station_ids=q.wmo_station_ids,
        source_ids=q.source_ids,
    )
    enhance(result, timezone=q.timezone, units=q.units)
    return ORJSONResponse(result)


@app.get(
    '/weather',
    operation_id='getWeather',
    summary='Hourly weather (including forecasts)',
    responses=common_responses,
)
async def weather(
    q: Annotated[WeatherParams, Query()],
) -> WeatherResponse:
    """
    Returns a list of hourly weather records (and/or forecasts) for the time
    range given by `date` and `last_date`.

    To set the location for which to retrieve records (and/or forecasts), you
    must supply both `lat` and `lon` _or_ one of `dwd_station_id`,
    `wmo_station_id`, or `source_id`.
    """
    result = await query.weather(
        ctx['pool'],
        date=q.date,
        last_date=q.last_date,
        lat=q.lat,
        lon=q.lon,
        max_dist=q.max_dist,
        dwd_station_ids=q.dwd_station_ids,
        wmo_station_ids=q.wmo_station_ids,
        source_ids=q.source_ids,
    )
    enhance(result, timezone=q.timezone, units=q.units)
    return ORJSONResponse(result)


@app.get(
    '/synop',
    operation_id='getSynop',
    summary='Raw SYNOP observations',
    tags=['Internals'],
    responses=common_responses,
)
async def synop(
    q: Annotated[SynopParams, Query()],
) -> SynopResponse:
    """
    Returns a list of ten-minutely SYNOP observations for the time range given
    by `date` and `last_date`. Note that Bright Sky only stores SYNOP
    observations from the past 30 hours.

    To set the weather station for which to retrieve records, you must supply
    one of `dwd_station_id`, `wmo_station_id`, or `source_id`. The `/synop`
    endpoint does not support `lat` and `lon`; use the
    [`/sources` endpoint](/operations/getSources) if you need to retrieve a
    SYNOP station ID close to a given location.

    SYNOP observations are stored as they were reported, which in particular
    implies that many parameters are only available at certain timestamps. For
    example, most stations report `sunshine_60` only on the full hour, and
    `sunshine_30` only at 30 minutes past the full hour (i.e. also not on the
    full hour). Check out the
    [`/current_weather` endpoint](/operations/getCurrentWeather) for an
    opinionated compilation of recent SYNOP records into a single "current
    weather" record.
    """
    result = await query.synop(
        ctx['pool'],
        date=q.date,
        last_date=q.last_date,
        dwd_station_ids=q.dwd_station_ids,
        wmo_station_ids=q.wmo_station_ids,
        source_ids=q.source_ids,
    )
    enhance(result, timezone=q.timezone, units=q.units)
    return ORJSONResponse(result)


class BytesORJSONResponse(ORJSONResponse):
    @staticmethod
    def encode_bytes(o):
        if isinstance(o, (bytes, memoryview)):
            return base64.b64encode(o).decode('ascii')
        raise TypeError

    def render(self, content: Any) -> bytes:
        return orjson.dumps(
            content,
            default=self.encode_bytes,
            option=orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY,
        )


@app.get(
    '/radar',
    operation_id='getRadar',
    summary='Radar',
    responses=common_responses,
)
async def radar(
    q: Annotated[RadarParams, Query()],
    request: Request,
    response: Response,
) -> RadarResponse:
    """
    Returns radar rainfall data with 1 km spatial and 5 minute temporal
    resolution, including a forecast for the next two hours.

    Radar data is recorded on a 1200 km (North-South) x 1100 km (East-West)
    grid, with each pixel representing 1 km². **That's quite a lot of data, so
    use `lat`/`lon` or `bbox` whenever you can (see below).** Past radar
    records are kept for six hours.

    Bright Sky can return the data in a few formats. Use the default
    `compressed` format if possible – this'll get you the fastest response
    times by far and reduce load on the server. If you have a small-ish
    bounding box (e.g. 250 x 250 pixels), using the `plain` format should be
    fine.

    ### Quickstart

    This request will get you radar data near Münster, reaching 200 km to the
    East/West/North/South, as a two-dimensional grid of integers:

    [`https://api.brightsky.dev/radar?lat=52&lon=7.6&format=plain`](https://api.brightsky.dev/radar?lat=52&lon=7.6&format=plain)

    ### Content

    * The grid is a polar stereographic projection of Germany and the regions
      bordering it. This is different from the mercator projection used for
      most consumer-facing maps like OpenStreetMap or Google Maps, and
      overlaying the radar data onto such a map without conversion
      (reprojection) will be inaccurate! Check out our [radar
      demo](https://brightsky.dev/demo/radar/) for an example of correctly
      reprojecting the radar data using OpenLayers. Alternatively, take a look
      at the `dwd:RV-Produkt` layer on the [DWD's open
      GeoServer](https://maps.dwd.de/geoserver/web/wicket/bookmarkable/org.geoserver.web.demo.MapPreviewPage)
      for ready-made tiles you can overlay onto a map.
    * The [proj-string](https://proj.org/en/9.2/usage/quickstart.html) for the
      grid projection is `+proj=stere +lat_0=90 +lat_ts=60 +lon_0=10 +a=6378137
      +b=6356752.3142451802 +no_defs +x_0=543196.83521776402
      +y_0=3622588.8619310018`. The radar pixels range from `-500` (left) to
      `1099500` (right) on the x-axis, and from `500` (top) to `-1199500`
      (bottom) on the y-axis, each radar pixel a size of `1000x1000` (1 km²).
    * The DWD data does not cover the whole grid! Many areas near the edges
      will always be `0`.
    * Values represent 0.01 mm / 5 min. I.e., if a pixel has a value of `45`,
      then 0.45 mm of precipitation fell in the corresponding square kilometer
      in the past five minutes.
    * The four corners of the grid are as follows:
      * Northwest: Latitude `55.86208711`, Longitude `1.463301510`
      * Northeast: Latitude `55.84543856`, Longitude `18.73161645`
      * Southeast: Latitude `45.68460578`, Longitude `16.58086935`
      * Southwest: Latitude `45.69642538`, Longitude `3.566994635`

    You can find details and more information in the [DWD's `RV product info`
    (German
    only)](https://www.dwd.de/DE/leistungen/radarprodukte/formatbeschreibung_rv.pdf?__blob=publicationFile&v=3).
    Below is an example visualization of the rainfall radar data taken from
    this document, using the correct projection and showing the radar coverage:

    ![image](https://github.com/jdemaeyer/brightsky/assets/10531844/09f9bb5f-088a-417e-8a0c-ea5a20fe0969)

    ### Code examples

    > The radar data is quite big (naively unpacking the default 25-frames
    > response into Python integer arrays will eat roughly 125 MB of memory),
    > so use `bbox` whenever you can.

    #### `compressed` format

    With Javascript using [`pako`](https://github.com/nodeca/pako):

    ```js
    fetch(
      'https://api.brightsky.dev/radar'
    ).then((resp) => resp.json()
    ).then((respData) => {
      const raw = respData.radar[0].precipitation_5;
      const compressed = Uint8Array.from(atob(raw), c => c.charCodeAt(0));
      const rawBytes = pako.inflate(compressed).buffer;
      const precipitation = new Uint16Array(rawBytes);
    });
    ```

    With Python using `numpy`:

    ```python
    import base64
    import zlib

    import numpy as np
    import requests

    resp = requests.get('https://api.brightsky.dev/radar')
    raw = resp.json()['radar'][0]['precipitation_5']
    raw_bytes = zlib.decompress(base64.b64decode(raw))

    data = np.frombuffer(
        raw_bytes,
        dtype='i2',
    ).reshape(
        # Adjust `1200` and `1100` to the height/width of your bbox
        (1200, 1100),
    )
    ```

    With Python using the standard library's `array`:
    ```python
    import array

    # [... load raw_bytes as above ...]

    data = array.array('H')
    data.frombytes(raw_bytes)
    data = [
        # Adjust `1200` and `1100` to the height/width of your bbox
        data[row*1100:(row+1)*1100]
        for row in range(1200)
    ]
    ```

    Simple plot using `matplotlib`:
    ```python
    import matplotlib.pyplot as plt

    # [... load data as above ...]

    plt.imshow(data, vmax=50)
    plt.show()
    ```

    #### `bytes` format

    Same as for `compressed`, but add `?format=bytes` to the URL and remove the
    call to `zlib.decompress`, using just `raw_bytes = base64.b64decode(raw)`
    instead.

    #### `plain` format

    This is obviously a lot simpler than the `compressed` format. It is,
    however, also a lot slower. Nonetheless, if you have a small-ish `bbox` the
    performance difference becomes manageable, so just using the `plain` format
    and not having to deal with unpacking logic can be a good option in this
    case.

    With Python:
    ```python
    import requests

    resp = requests.get('https://api.brightsky.dev/radar?format=plain')
    data = resp.json()['radar'][0]['precipitation_5']
    ```

    ### Additional resources

    * [Source for our radar demo, including reprojecton via OpenLayers](https://github.com/jdemaeyer/brightsky/blob/master/docs/demo/radar/index.html)
    * [Raw data on the Open Data Server](https://opendata.dwd.de/weather/radar/composite/rv/)
    * [Details on the `RV` product (German)](https://www.dwd.de/DE/leistungen/radarprodukte/formatbeschreibung_rv.pdf?__blob=publicationFile&v=3)
    * [Visualization of current rainfall radar](https://www.dwd.de/DE/leistungen/radarbild_film/radarbild_film.html)
    * [General info on DWD radar products (German)](https://www.dwd.de/DE/leistungen/radarprodukte/radarprodukte.html)
    * [Radar status (German)](https://www.dwd.de/DE/leistungen/radarniederschlag/rn_info/home_freie_radarstatus_kartendaten.html?nn=16102)
    * [DWD notifications for radar products (German)](https://www.dwd.de/DE/leistungen/radolan/radolan_info/radolan_informationen.html?nn=16102)
    """
    if q.format == 'compressed':
        # Prevent traefik from gzipping the pre-compressed content
        response.headers['Content-Encoding'] = 'identity'
    else:
        allowed_encodings = ['br', 'zstd', 'gzip']
        accepted_encoding = request.headers.get('accept-encoding', '')
        if not any(e in accepted_encoding for e in allowed_encodings):
            raise HTTPException(
                status_code=400,
                detail=(
                    "Requests to the radar endpoint with format 'plain' or "
                    "'bytes' must accept br, zstd, or gzip encoding"
                ),
            )
    result = await query.radar(
        ctx['pool'],
        date=q.date,
        last_date=q.last_date,
        lat=q.lat,
        lon=q.lon,
        distance=q.distance,
        fmt=q.format,
        bbox=q.bbox,
    )
    enhance(result, timezone=q.timezone)
    return BytesORJSONResponse(result)


@app.get(
    '/alerts',
    operation_id='getAlerts',
    summary='Alerts',
    responses=common_responses,
)
async def alerts(
    q: Annotated[AlertsParams, Query()],
) -> AlertsResponse:
    """
    Returns a list of weather alerts for the given location, or all weather
    alerts if no location given.

    If you supply either `warn_cell_id` or both `lat` and `lon`, Bright Sky
    will return additional information on that cell in the `location` field.
    Warn cell IDs are municipality (_Gemeinde_) cell IDs.

    ### Notes

    * The DWD divides Germany's area into roughly 11,000 "cells" based on
      municipalities (_Gemeinden_), and issues warnings for each of these
      cells. Most warnings apply to many cells.
    * Bright Sky will supply information on the cell that a given lat/lon
      belongs to in the `location` field.
    * There is also a set of roughly 400 cells based on districts
      (_Landkreise_) that is not supported by Bright Sky.
    * The complete list of cells can be found on the DWD GeoServer (see below).

    ### Additional resources

    * [Live demo of a simple interactive alerts map](https://brightsky.dev/demo/alerts/)
    * [Source for alerts map demo](https://github.com/jdemaeyer/brightsky/blob/master/docs/demo/alerts/index.html)
    * [Map view of all current alerts by the DWD](https://www.dwd.de/DE/wetter/warnungen_gemeinden/warnWetter_node.html)
    * [List of municipality cells](https://maps.dwd.de/geoserver/wfs?SERVICE=WFS&VERSION=2.0.0&REQUEST=GetFeature&TYPENAMES=Warngebiete_Gemeinden&OUTPUTFORMAT=json)
    * [List of district cells (*not used by Bright Sky!*)](https://maps.dwd.de/geoserver/wfs?SERVICE=WFS&VERSION=2.0.0&REQUEST=GetFeature&TYPENAMES=Warngebiete_Kreise&OUTPUTFORMAT=json)
    * [Raw data on the Open Data Server](https://opendata.dwd.de/weather/alerts/cap/COMMUNEUNION_DWD_STAT/)
    * [DWD Documentation on alert fields and their allowed contents (English)](https://www.dwd.de/DE/leistungen/opendata/help/warnungen/cap_dwd_profile_en_pdf_2_1_13.pdf?__blob=publicationFile&v=3)
    * [DWD Documentation on alert fields and their allowed contents (German)](https://www.dwd.de/DE/leistungen/opendata/help/warnungen/cap_dwd_profile_de_pdf_2_1_13.pdf?__blob=publicationFile&v=3)
    """
    result = await query.alerts(
        ctx['pool'],
        lat=q.lat,
        lon=q.lon,
        warn_cell_id=q.warn_cell_id,
    )
    enhance(result, timezone=q.timezone)
    return ORJSONResponse(result)



import datetime
import re
from contextlib import suppress
from typing import Annotated, Literal

from dateutil.tz import gettz
from fastapi import Query
from pydantic import BaseModel, Field, field_validator, model_validator


class LatLon(BaseModel):
    lat: float = Field(
        ge=-90,
        le=90,
        default=None,
        description="Latitude in decimal degrees.",
        examples=[
            52.52,
            51.55,
        ],
    )
    lon: float = Field(
        ge=-180,
        le=180,
        default=None,
        description="Longitude in decimal degrees.",
        examples=[
            13.4,
            9.9,
        ],
    )

    @model_validator(mode='after')
    def validate_both_or_no_lat_lon(self):
        if (self.lat is None) != (self.lon is None):
            raise ValueError("Supply both lat and lon, or none of the two")
        return self


class MaxDist(BaseModel):
    max_dist: int = Field(
        ge=0,
        le=500000,
        default=50000,
        description="Maximum distance of record location from the location given by `lat` and `lon`, in meters. Only has an effect when using `lat` and `lon`.",  # noqa
        examples=[
            10000,
        ],
    )


def _split(value, converter=str, sep=','):
    if len(value) > 1:
        # e.g. ?param=1&param=2
        return value
    # e.g. ?param=1,2
    def _convert(x):
        with suppress(Exception):
            return converter(x)
        raise ValueError(f"'{x}' is not a valid {converter.__name__}")
    return [_convert(x) for x in value[0].split(sep) if x]


class StationIDs(BaseModel):
    dwd_station_ids: Annotated[
        list[str],
        Query(alias='dwd_station_id'),
    ] = Field(
        default=None,
        description="DWD station ID, typically five alphanumeric characters. You can supply multiple station IDs separated by commas, ordered from highest to lowest priority.",  # noqa
        examples=[
            "01766",
            "00420,00053,00400",
        ],
    )
    wmo_station_ids: Annotated[
        list[str],
        Query(alias='wmo_station_id'),
    ] = Field(
        default=None,
        description="WMO station ID, typically five alphanumeric characters. You can supply multiple station IDs separated by commas, ordered from highest to lowest priority.",  # noqa
        examples=[
            "10315",
            "G005,F451,10389",
        ],
    )

    @field_validator('dwd_station_ids', mode='before')
    @classmethod
    def validate_dwd_station_ids(cls, value):
        return _split(value)

    @field_validator('wmo_station_ids', mode='before')
    @classmethod
    def validate_wmo_station_ids(cls, value):
        return _split(value)


class SourceIDs(BaseModel):
    source_ids: Annotated[
        list[int],
        Query(alias='source_id'),
    ] = Field(
        default=None,
        description="Bright Sky source ID, as retrieved from the [`/sources` endpoint](/operations/getSources). You can supply multiple source IDs separated by commas, ordered from highest to lowest priority.",  # noqa
        examples=[
            "1234",
            "1234,2345",
        ],
    )

    @field_validator('source_ids', mode='before')
    @classmethod
    def validate_source_ids(cls, value):
        return _split(value, converter=int)


class DateRange(BaseModel):
    date: datetime.datetime = Field(
        description="Timestamp of first weather record (or forecast) to retrieve, in ISO 8601 format. May contain time and/or UTC offset.",  # noqa
        examples=[
            "2023-08-07",
            "2023-08-07T08:00+02:00",
        ],
    )
    last_date: datetime.datetime = Field(
        default=None,
        description="Timestamp of last weather record (or forecast) to retrieve, in ISO 8601 format. Will default to `date + 1 day`.",  # noqa
        examples=[
            "2023-08-08",
            "2023-08-07T23:00+02:00",
        ],
    )

    @field_validator('date', 'last_date', mode='before')
    def fix_unescaped_offset(cls, value):
        # Handle " 02:00" (i.e. space in raw URL) instead of "%2B02:00"
        return re.sub(r' (\d{2}:\d{2})$', r'+\1', value)

    @model_validator(mode='after')
    def set_default_last_date(self):
        if self.last_date is None:
            self.last_date = self.date + datetime.timedelta(days=1)
        return self

    @model_validator(mode='after')
    def ensure_tzinfo(self):
        default = getattr(self, 'timezone', None) or datetime.UTC
        if not self.date.tzinfo:
            self.date = self.date.replace(tzinfo=default)
        if not self.last_date.tzinfo:
            self.last_date = self.last_date.replace(tzinfo=default)
        return self


class Timezone(BaseModel):
    timezone: Annotated[
        str,
        Query(alias='tz'),
    ] = Field(
        default=None,
        description="Timezone in which record timestamps will be presented, as <a href=\"https://en.wikipedia.org/wiki/List_of_tz_database_time_zones\">tz database name</a>. Will also be used as timezone when parsing `date` and `last_date`, unless these have explicit UTC offsets. If omitted but `date` has an explicit UTC offset, that offset will be used as timezone. Otherwise will default to UTC.",  # noqa
        examples=[
            "Europe/Berlin",
            "Australia/Darwin",
            "Etc/UTC",
        ],
    )

    @field_validator('timezone', mode='after')
    @classmethod
    def validate_timezone(cls, value):
        if value and not gettz(value):
            raise ValueError(f"Unknown timezone: {value}")
        return gettz(value)

    @model_validator(mode='after')
    def set_default_timezone_from_date(self):
        if self.timezone:
            return self
        if getattr(self, 'date', None) and self.date.tzinfo != datetime.UTC:
            self.timezone = self.date.tzinfo
        return self


class Units(BaseModel):
    units: Literal['dwd', 'si'] = Field(
        default='dwd',
        description="""\
Physical units in which meteorological parameters will be returned. Set to `si` to use <a href="https://en.wikipedia.org/wiki/International_System_of_Units">SI units</a> (except for precipitation, which is always measured in millimeters). The default `dwd` option uses a set of units that is more common in meteorological applications and civil use:
<table>
  <tr><td></td><td>DWD</td><td>SI</td></tr>
  <tr><td>Cloud cover</td><td>%</td><td>%</td></tr>
  <tr><td>Dew point</td><td>°C</td><td>K</td></tr>
  <tr><td>Precipitation</td><td>mm</td><td><s>kg / m²</s> <strong>mm</strong></td></tr>
  <tr><td>Precipitation probability</td><td>%</td><td>%</td></tr>
  <tr><td>Pressure</td><td>hPa</td><td>Pa</td></tr>
  <tr><td>Relative humidity</td><td>%</td><td>%</td></tr>
  <tr><td>Solar irradiation</td><td>kWh / m²</td><td>J / m²</td></tr>
  <tr><td>Sunshine</td><td>min</td><td>s</td></tr>
  <tr><td>Temperature</td><td>°C</td><td>K</td></tr>
  <tr><td>Visibility</td><td>m</td><td>m</td></tr>
  <tr><td>Wind (gust) direction</td><td>°</td><td>°</td></tr>
  <tr><td>Wind (gust) speed</td><td>km / h</td><td>m / s</td></tr>
</table>""",  # noqa
    )


class WarnCell(BaseModel):
    warn_cell_id: int = Field(
        default=None,
        description="Municipality warn cell ID.",
        examples=[
            803159016,
            705515101,
        ],
    )


class RadarDateRange(BaseModel):
    date: datetime.datetime = Field(
        default=None,
        description="Timestamp of first record to retrieve, in ISO 8601 format. May contain time and/or UTC offset. (_Defaults to 1 hour before latest measurement._)",  # noqa
        examples=[
            "2023-08-07",
            "2023-08-07T19:00+02:00",
        ],
    )
    last_date: datetime.datetime = Field(
        default=None,
        description="Timestamp of last record to retrieve, in ISO 8601 format. May contain time and/or UTC offset. (_Defaults to 2 hours after `date`._)",  # noqa
        examples=[
            "2023-08-08",
            "2023-08-07T23:00+02:00",
        ],
    )

    @field_validator('date', 'last_date', mode='before')
    def fix_unescaped_offset(cls, value):
        # Handle " 02:00" (i.e. space in raw URL) instead of "%2B02:00"
        return re.sub(r' (\d{2}:\d{2})$', r'+\1', value)

    @model_validator(mode='after')
    def set_default_last_date(self):
        if self.last_date is None and self.date is not None:
            self.last_date = self.date + datetime.timedelta(hours=2)
        return self

    @field_validator('date', 'last_date', mode='after')
    @classmethod
    def ensure_tzinfo(cls, value):
        if value is None or value.tzinfo:
            return value
        return value.replace(tzinfo=datetime.UTC)


class RadarBoundingBox(BaseModel):
    bbox: list[int] = Field(
        default=None,
        description="Bounding box (top, left, bottom, right) **in pixels**, edges are inclusive. (_Defaults to full 1200x1100 grid._)",  # noqa
        examples=[
            "100,100,300,300",
        ],
    )
    distance: int = Field(
        default=200000,
        description="Alternative way to set a bounding box, must be used together with `lat` and `lon`. Data will reach `distance` meters to each side of this location, but is possibly cut off at the edges of the radar grid.",  # noqa
        examples=[
            100000,
        ],
    )

    @field_validator('bbox', mode='before')
    @classmethod
    def validate_bbox(cls, value):
        bbox = _split(value, converter=int)
        if len(bbox) != 4:
            raise ValueError(
                "The 'bbox' parameter must be a comma-separated list of four "
                "integers: top, left, bottom, right (edges are inclusive)."
            )
        return bbox


class RadarFormat(BaseModel):
    format: Literal['compressed', 'bytes', 'plain'] = Field(
        default='compressed',
        description="""
Determines how the precipitation data is encoded into the `precipitation_5` field:
* `compressed`: base64-encoded, zlib-compressed bytestring of 2-byte integers
* `bytes`: base64-encoded bytestring of 2-byte integers
* `plain`: Nested array of integers
        """,  # noqa
    )


class SourcesParams(
    Timezone,
    SourceIDs,
    StationIDs,
    MaxDist,
    LatLon,
):
    @model_validator(mode='after')
    def validate_at_least_one_option(self):
        if self.lat is not None and self.lon is not None:
            return self
        elif any([
            self.dwd_station_ids,
            self.wmo_station_ids,
            self.source_ids,
        ]):
            return self
        raise ValueError(
            "Please supply lat & lon, or dwd_station_id, or wmo_station_id, "
            "or source_id",
        )


class CurrentWeatherParams(
    Units,
    SourcesParams,
):
    pass


class WeatherParams(
    Units,
    SourcesParams,
    DateRange,
):
    pass


class SynopParams(
    Units,
    Timezone,
    SourceIDs,
    StationIDs,
    DateRange,
):
    @model_validator(mode='after')
    def validate_at_least_one_option(self):
        if any([
            self.dwd_station_ids,
            self.wmo_station_ids,
            self.source_ids,
        ]):
            return self
        raise ValueError(
            "Please dwd_station_id or wmo_station_id or source_id",
        )


class RadarParams(
    Timezone,
    RadarFormat,
    RadarDateRange,
    LatLon,
    RadarBoundingBox,
):
    pass


class AlertsParams(
    Timezone,
    WarnCell,
    LatLon,
):
    pass



import datetime
import json
import os
import tempfile
from functools import cached_property

import numpy as np
import requests
from isal import isal_zlib as zlib
from pyproj import CRS, Transformer
from shapely import MultiPolygon, STRtree, Point

from brightsky.settings import settings
from brightsky.utils import USER_AGENT


class NoData(LookupError):
    pass


class PgParams(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.map = {}

    def __getitem__(self, key):
        return f'${self.map.setdefault(key, len(self.map) + 1)}'

    def get_params(self):
        get_value = super().__getitem__
        return tuple(get_value(k) for k in self.map)


def topg(query, params):
    p = PgParams(params)
    return query.format_map(p), p.get_params()


def make_dicts(rows):
    return [dict(row) for row in rows]


async def weather(
    conn,
    date,
    last_date,
    lat=None,
    lon=None,
    max_dist=50000,
    dwd_station_ids=None,
    wmo_station_ids=None,
    source_ids=None,
):
    sources_data = await sources(
        conn,
        lat=lat,
        lon=lon,
        max_dist=max_dist,
        dwd_station_ids=dwd_station_ids,
        wmo_station_ids=wmo_station_ids,
        source_ids=source_ids,
        observation_types=['historical', 'current', 'forecast'],
        date=date,
        last_date=last_date,
    )
    sources_rows = sources_data['sources']
    primary_source_ids = {}
    for row in sources_rows:
        primary_source_ids.setdefault(row['observation_type'], row['id'])
    primary_source_ids = list(primary_source_ids.values())
    weather_rows = await _weather(conn, date, last_date, primary_source_ids)
    source_ids = [row['id'] for row in sources_rows]
    if len(weather_rows) < int((last_date - date).total_seconds()) // 3600:
        weather_rows = await _weather(conn, date, last_date, source_ids)
    await _fill_missing_fields(
        conn,
        weather_rows,
        date,
        last_date,
        source_ids,
        True,
    )
    await _fill_missing_fields(
        conn,
        weather_rows,
        date,
        last_date,
        source_ids,
        False,
    )
    used_source_ids = {row['source_id'] for row in weather_rows}
    used_source_ids.update(
        source_id
        for row in weather_rows
        for source_id in row.get('fallback_source_ids', {}).values()
    )
    return {
        'weather': weather_rows,
        'sources': [s for s in sources_rows if s['id'] in used_source_ids],
    }


async def _weather(
    conn,
    date,
    last_date,
    source_ids,
    not_null=None,
    not_null_or=False,
):
    params = {
        'date': date,
        'last_date': last_date,
        'source_ids': source_ids,
    }
    where = "timestamp BETWEEN {date} AND {last_date}"
    order_by = "timestamp"
    where += " AND source_id = ANY({source_ids}::int[])"
    order_by += ", array_position({source_ids}::int[], source_id)"
    if not_null:
        glue = ' OR ' if not_null_or else ' AND '
        constraint = glue.join(f"{x} IS NOT NULL" for x in not_null)
        where += f" AND ({constraint})"
    sql = f"""
        SELECT DISTINCT ON (timestamp) *
        FROM weather
        WHERE {where}
        ORDER BY {order_by}
    """
    sql, params = topg(sql, params)
    rows = await conn.fetch(sql, *params)
    return make_dicts(rows)


IGNORED_MISSING_FIELDS = {
    # Not available in MOSMIX
    'relative_humidity',
    'wind_gust_direction',
    # Not available in recent and historical measurements
    'precipitation_probability',
    'precipitation_probability_6h',
    # Not measured at many stations
    'solar',
}


async def _fill_missing_fields(
    conn,
    weather_rows,
    date,
    last_date,
    source_ids,
    partial,
):
    incomplete_rows = []
    missing_fields = set()
    for row in weather_rows:
        missing_row_fields = set(k for k, v in row.items() if v is None)
        relevant_fields = missing_row_fields.difference(IGNORED_MISSING_FIELDS)
        if relevant_fields:
            incomplete_rows.append((row, missing_row_fields))
            missing_fields.update(relevant_fields)
    if missing_fields:
        min_date = incomplete_rows[0][0]['timestamp']
        max_date = incomplete_rows[-1][0]['timestamp']
        fallback_rows = {
            row['timestamp']: row
            for row in await _weather(
                conn,
                min_date,
                max_date,
                source_ids,
                not_null=missing_fields,
                not_null_or=partial,
            )
        }
        for row, fields in incomplete_rows:
            fallback_row = fallback_rows.get(row['timestamp'])
            if fallback_row:
                row.setdefault('fallback_source_ids', {})
                for f in fields:
                    if fallback_row[f] is None:
                        continue
                    row[f] = fallback_row[f]
                    row['fallback_source_ids'][f] = fallback_row['source_id']


async def current_weather(
    conn,
    lat=None,
    lon=None,
    max_dist=50000,
    dwd_station_ids=None,
    wmo_station_ids=None,
    source_ids=None,
    fallback=True,
):
    sources_data = await sources(
        conn,
        lat=lat,
        lon=lon,
        max_dist=max_dist,
        dwd_station_ids=dwd_station_ids,
        wmo_station_ids=wmo_station_ids,
        source_ids=source_ids,
        observation_types=['synop'],
    )
    sources_rows = sources_data['sources']
    source_ids = [row['id'] for row in sources_rows]
    weather = await _current_weather(conn, source_ids)
    if not weather:
        raise NoData(
            "Could not find current weather for your location criteria",
        )
    used_source_ids = [weather['source_id']]
    if fallback:
        missing_fields = [k for k, v in weather.items() if v is None]
        fallback_weather = await _current_weather(
            conn,
            source_ids,
            not_null=missing_fields,
        )
        if fallback_weather:
            weather.update({k: fallback_weather[k] for k in missing_fields})
            weather['fallback_source_ids'] = {
                field: fallback_weather['source_id']
                for field in missing_fields}
            used_source_ids.append(fallback_weather['source_id'])
    return {
        'weather': weather,
        'sources': [s for s in sources_rows if s['id'] in used_source_ids],
    }


async def _current_weather(conn, source_ids, not_null=None, partial=False):
    params = {
        'source_ids': source_ids,
    }
    where = "source_id = ANY({source_ids}::int[])"
    if not_null:
        glue = ' OR ' if partial else ' AND '
        extra = glue.join(f"{element} IS NOT NULL" for element in not_null)
        where += f'AND ({extra})'
    sql = f"""
        SELECT *
        FROM current_weather
        WHERE {where}
        ORDER BY array_position({{source_ids}}::int[], source_id)
        LIMIT 1
    """
    sql, params = topg(sql, params)
    row = await conn.fetchrow(sql, *params)
    if not row:
        if not_null and not partial:
            return await _current_weather(
                conn,
                source_ids,
                not_null=not_null,
                partial=True,
            )
        return {}
    return dict(row)


async def synop(
    conn,
    date,
    last_date,
    dwd_station_ids=None,
    wmo_station_ids=None,
    source_ids=None,
):
    sources_data = await sources(
        conn,
        dwd_station_ids=dwd_station_ids,
        wmo_station_ids=wmo_station_ids,
        source_ids=source_ids,
        observation_types=['synop'],
    )
    sources_rows = sources_data['sources']
    source_ids = [row['id'] for row in sources_rows]
    sql = """
        SELECT *
        FROM synop
        WHERE
            timestamp BETWEEN {date} AND {last_date} AND
            source_id = ANY({source_ids}::int[])
        ORDER BY timestamp
        """
    params = {
        'date': date,
        'last_date': last_date,
        'source_ids': source_ids,
    }
    sql, params = topg(sql, params)
    rows = await conn.fetch(sql, *params)
    return {
        'weather': make_dicts(rows),
        'sources': make_dicts(sources_rows),
    }


async def radar(
    conn,
    date=None,
    last_date=None,
    lat=None,
    lon=None,
    distance=200000,
    fmt='compressed',
    bbox=None,
):
    extra = {}
    if not date:
        date = await conn.fetchval(
            "SELECT MAX(timestamp) - '3 hours'::interval FROM radar"
        )
    if not last_date:
        last_date = date + datetime.timedelta(hours=2)
    if lat is not None and lon is not None:
        x, y = _transformer.to_xy(lat, lon)
        if not -0.5 <= x <= 1099.5 or not -0.5 <= y <= 1199.5:
            raise NoData("lat/lon lies outside the radar data range")
        center_x = int(round(x))
        center_y = int(round(y))
        pixels = distance // 1000
        bbox = (
            max(center_y - pixels, 0),
            max(center_x - pixels, 0),
            min(center_y + pixels, 1199),
            min(center_x + pixels, 1099),
        )
        extra['bbox'] = bbox
        extra['latlon_position'] = {
            'x': round(x - bbox[1], 3),
            'y': round(y - bbox[0], 3),
        }
    sql = """
        SELECT *
        FROM radar
        WHERE timestamp BETWEEN {date} AND {last_date}
        ORDER BY timestamp
        """
    params = {
        'date': date,
        'last_date': last_date,
    }
    sql, params = topg(sql, params)
    rows = make_dicts(await conn.fetch(sql, *params))
    if fmt == 'plain':
        for row in rows:
            row['precipitation_5'] = _load_radar(row['precipitation_5'], bbox)
    elif fmt == 'bytes':
        for row in rows:
            row['precipitation_5'] = memoryview(
                _load_radar(row['precipitation_5'], bbox),
            )
    elif fmt == 'compressed' and bbox:
        for row in rows:
            row['precipitation_5'] = zlib.compress(
                _load_radar(row['precipitation_5'], bbox),
            )
    elif fmt != 'compressed':
        raise ValueError(f"Unknown format: '{fmt}'")
    return {
        'radar': rows,
        'geometry': _transformer.bbox_to_geometry(bbox),
        **extra,
    }


def _load_radar(raw, bbox, width=1100, height=1200):
    precip = np.frombuffer(
        zlib.decompress(raw),
        dtype='i2',
    ).reshape((height, width))
    if bbox:
        top, left, bottom, right = bbox
        precip = precip[top:bottom+1, left:right+1]
        # Arrays must be C-contiguous for orjson and zlib
        precip = np.ascontiguousarray(precip).reshape(precip.shape)
    return precip


class RadarCoordinatesTransformer:

    PROJ_STR = (
        "+proj=stere +lat_0=90 +lat_ts=60 +lon_0=10 +a=6378137 "
        "+b=6356752.3142451802 +no_defs +x_0=543196.83521776402 "
        "+y_0=3622588.8619310018"
    )

    @cached_property
    def de1200(self):
        return CRS.from_proj4(self.PROJ_STR)

    @cached_property
    def wgs84_to_de1200(self):
        return Transformer.from_crs(4326, self.de1200)

    @cached_property
    def de1200_to_wgs84(self):
        return Transformer.from_crs(self.de1200, 4326)

    def to_xy(self, lat, lon):
        x, y = self.wgs84_to_de1200.transform(lat, lon)
        return round(x) / 1000, -round(y) / 1000

    def to_latlon(self, x, y):
        lat, lon = self.de1200_to_wgs84.transform(x * 1000, -y * 1000)
        return round(lat, 5), round(lon, 5)

    def to_lonlat(self, x, y):
        return tuple(reversed(self.to_latlon(x, y)))

    def bbox_to_geometry(self, bbox):
        if not bbox:
            bbox = (0, 0, 1199, 1099)
        top, left, bottom, right = bbox
        return {
            'type': 'Polygon',
            'coordinates': [
                self.to_lonlat(left - .5, top - .5),
                self.to_lonlat(left - .5, bottom + .5),
                self.to_lonlat(right + .5, bottom + .5),
                self.to_lonlat(right + .5, top - .5),
            ],
        }


_transformer = RadarCoordinatesTransformer()


async def alerts(
    conn,
    lat=None,
    lon=None,
    warn_cell_id=None,
):
    if lat is not None and lon is not None:
        meta = _warn_cells.find(lat, lon)
    elif warn_cell_id is not None:
        try:
            meta = _warn_cells.get_meta(warn_cell_id)
        except KeyError:
            raise NoData(
                "Unknown warn_cell_id, please use commune (Gemeinden), not "
                "district (Landkreis) ids"
            )
    else:
        sql = """
            SELECT *
            FROM alerts
            JOIN (
                SELECT alert_id, array_agg(warn_cell_id) as warn_cell_ids
                FROM alert_cells
                GROUP BY alert_id
            ) cells ON alerts.id = cells.alert_id
            ORDER BY severity DESC
        """
        rows = await conn.fetch(sql)
        return {'alerts': make_dicts(rows)}
    sql = """
        SELECT *
        FROM alerts
        WHERE id IN (
            SELECT alert_id
            FROM alert_cells
            WHERE warn_cell_id = {warn_cell_id}
        )
        ORDER BY severity DESC
        """
    params = {
        'warn_cell_id': meta['warn_cell_id'],
    }
    sql, params = topg(sql, params)
    rows = await conn.fetch(sql, *params)
    return {
        'alerts': make_dicts(rows),
        'location': meta,
    }


class WarnCellManager:

    CELLS_CACHE_PATH = os.path.join(tempfile.gettempdir(), 'alert_cells.json')

    @cached_property
    def tree(self):
        self.cell_meta = {}
        self.cell_meta_by_id = {}
        for f in self.get_cell_data()['features']:
            polygons = [
                # shell, holes
                (c[0], c[1:])
                for c in f['geometry']['coordinates']
            ]
            p = MultiPolygon(polygons)
            meta = {
                'warn_cell_id': f['properties']['WARNCELLID'],
                'name': f['properties']['NAME'],
                'name_short': f['properties']['KURZNAME'],
                'district': f['properties']['KREIS'],
                'state': f['properties']['BUNDESLAND'],
                'state_short': f['properties']['BL_KUERZEL'],
            }
            self.cell_meta[p] = meta
            self.cell_meta_by_id[meta['warn_cell_id']] = meta
        return STRtree(list(self.cell_meta.keys()))

    def get_cell_data(self):
        path = self.CELLS_CACHE_PATH
        if not os.path.isfile(path):
            resp = requests.get(
                settings.WARN_CELLS_URL,
                headers={'User-Agent': USER_AGENT},
            )
            with open(path, 'wb') as f:
                f.write(resp.content)
        with open(path) as f:
            return json.load(f)

    def find(self, lat, lon):
        p = Point(lon, lat)
        cell = self.tree.geometries[self.tree.nearest(p)]
        if cell.distance(p) > 0.01:
            raise NoData("Requested position is not covered by the DWD")
        return self.cell_meta[cell]

    def get_meta(self, warn_cell_id):
        # Make sure cells have been parsed
        self.tree
        return self.cell_meta_by_id[warn_cell_id]


_warn_cells = WarnCellManager()


async def sources(
    conn,
    lat=None,
    lon=None,
    max_dist=50000,
    dwd_station_ids=None,
    wmo_station_ids=None,
    source_ids=None,
    observation_types=None,
    ignore_type=False,
    date=None,
    last_date=None,
):
    select = "*"
    order_by = "observation_type"
    params = {
        'lat': lat,
        'lon': lon,
        'max_dist': max_dist,
        'dwd_station_ids': dwd_station_ids,
        'wmo_station_ids': wmo_station_ids,
        'source_ids': source_ids,
        'observation_types': observation_types,
        'date': date,
        'last_date': last_date,
    }
    if source_ids:
        where = "id = ANY({source_ids}::int[])"
        order_by = "array_position({source_ids}, id), observation_type"
    elif dwd_station_ids:
        where = "dwd_station_id = ANY({dwd_station_ids}::text[])"
        order_by = """
            array_position({dwd_station_ids}, dwd_station_id::text),
            observation_type
        """
    elif wmo_station_ids:
        where = "wmo_station_id = ANY({wmo_station_ids}::text[])"
        order_by = """
            array_position({wmo_station_ids}, wmo_station_id::text),
            observation_type
        """
    elif (lat is not None and lon is not None):
        distance = """
            earth_distance(ll_to_earth({lat}, {lon}), ll_to_earth(lat, lon))
        """
        select += f", round({distance}) AS distance"
        where = f"""
            earth_box(
                ll_to_earth({{lat}}, {{lon}}),
                {{max_dist}}
            ) @> ll_to_earth(lat, lon) AND
            {distance} < {{max_dist}}
        """
        if ignore_type:
            order_by = "distance"
        else:
            order_by += ", distance"
    else:
        raise ValueError(
            "Please supply lat & lon, or dwd_station_ids, or wmo_station_ids, "
            "or source_ids",
        )
    if observation_types:
        where += " AND observation_type = ANY({observation_types}::observation_type[])"  # noqa
    if date is not None:
        where += " AND last_record >= {date}"
    if last_date is not None:
        where += " AND first_record <= {last_date}"
    sql = f"""
        SELECT {select}
        FROM sources
        WHERE {where}
        ORDER BY {order_by}
    """
    sql, params = topg(sql, params)
    rows = await conn.fetch(sql, *params)
    if not rows:
        raise NoData("No sources match your criteria")
    return {'sources': make_dicts(rows)}



from dwdparse.units import convert_record

from brightsky.settings import settings
from brightsky.utils import daytime


def enhance(result, timezone=None, units='si'):
    if 'sources' in result:
        enhance_sources(result['sources'], timezone=timezone)
    if 'weather' in result:
        source_map = {s['id']: s for s in result['sources']}
        enhance_records(
            result['weather'],
            source_map,
            timezone=timezone,
            units=units,
        )
    if 'radar' in result:
        enhance_radar(result['radar'], timezone=timezone)
    if 'alerts' in result:
        enhance_alerts(result['alerts'], timezone=timezone)


def enhance_records(records, source_map, timezone=None, units='si'):
    if not isinstance(records, list):
        records = [records]
    for record in records:
        record['icon'] = get_icon(record, source_map)
        if units != 'si':
            convert_record(record, units)
        process_timestamp(record, 'timestamp', timezone)


def enhance_sources(sources, timezone=None):
    for source in sources:
        process_timestamp(source, 'first_record', timezone)
        process_timestamp(source, 'last_record', timezone)


def enhance_radar(radar, timezone=None):
    for record in radar:
        process_timestamp(record, 'timestamp', timezone)


def enhance_alerts(alerts, timezone=None):
    for alert in alerts:
        for key in ['effective', 'onset', 'expires']:
            process_timestamp(alert, key, timezone)


def process_timestamp(o, key, timezone):
    if not timezone:
        return
    elif not o[key]:
        return
    o[key] = o[key].astimezone(timezone)


def get_icon(record, source_map):
    if record['condition'] in (
            'fog', 'sleet', 'snow', 'hail', 'thunderstorm'):
        return record['condition']
    try:
        precipitation = record['precipitation']
    except KeyError:
        precipitation = record['precipitation_10']
    try:
        wind_speed = record['wind_speed']
    except KeyError:
        wind_speed = record['wind_speed_10']
    # Don't show 'rain' icon for little precipitation, and do show 'rain'
    # icon when condition is None but there is significant precipitation
    is_rainy = (
        record['condition'] == 'rain' and precipitation is None) or (
        (precipitation or 0) > settings.ICON_RAIN_THRESHOLD)
    if is_rainy:
        return 'rain'
    elif (wind_speed or 0) > settings.ICON_WIND_THRESHOLD:
        return 'wind'
    elif (record['cloud_cover'] or 0) >= settings.ICON_CLOUDY_THRESHOLD:
        return 'cloudy'
    source = source_map[record['source_id']]
    daytime_str = daytime(source['lat'], source['lon'], record['timestamp'])
    if (record['cloud_cover'] or 0) >= settings.ICON_PARTLY_CLOUDY_THRESHOLD:
        return f'partly-cloudy-{daytime_str}'
    return f'clear-{daytime_str}'
